/*
 *    Copyright 2006 Intel Corporation
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <dtn-config.h>
#endif

#include "BPQBlock.h"
#include "Bundle.h"
#include "BundleProtocol.h"
#include "SDNV.h"

namespace dtn {

// Setup our logging information
static const char* LOG = "/dtn/bundle/extblock/bpq";

BPQBlock::BPQBlock(Bundle* bundle)
{
    log_info_p(LOG, "BPQBlock::constructor()");

    if( bundle->recv_blocks().
        has_block(BundleProtocol::QUERY_EXTENSION_BLOCK) ) {

        log_debug_p(LOG, "BPQBlock found in Recv Block Vec => created remotly");
        initialise( const_cast<BlockInfo*> (bundle->recv_blocks().
                    find_block(BundleProtocol::QUERY_EXTENSION_BLOCK)) );

    } else if( bundle->api_blocks()->
               has_block(BundleProtocol::QUERY_EXTENSION_BLOCK) ) {

        log_debug_p(LOG, "BPQBlock found in API Block Vec => created locally");
        initialise( const_cast<BlockInfo*> (bundle->api_blocks()->
                    find_block(BundleProtocol::QUERY_EXTENSION_BLOCK)) );

    } else {
        log_err_p(LOG, "BPQ Block not found in bundle");
    }

    log_info_p(LOG, "leaving constructor");
}

BPQBlock::BPQBlock(BlockInfo* block)
{
    log_info_p(LOG, "BPQBlock::constructor()");

    ASSERT( block->type() == BundleProtocol::QUERY_EXTENSION_BLOCK);

    initialise(block);

    log_info_p(LOG, "leaving constructor");
}

BPQBlock::~BPQBlock()
{
    log_info_p(LOG, "BPQBlock: destructor");
    if ( query_val_ != NULL ){
        free(query_val_);
        query_val_ = NULL;
    }
}
/*
int 
BPQBlock::format(char* buf, size_t sz) const
{
    if ( kind_ == KIND_QUERY ) {
        return snprintf (buf, sz, "BPQ Query [%s] Matching Rule [%d]",
                         query_val_,
                         matching_rule_);
    } else if ( kind_ == KIND_RESPONSE ) {
        return snprintf (buf, sz, "BPQ Response [%s] Matching Rule [%d]",
                         query_val_,
                         matching_rule_);
    } else
        return snprintf (buf, sz, "INVALID BPQ KIND [%d]", kind_);
    }
}

void
BPQBlock::format_verbose(oasys::StringBuffer* buf) const
{
    if ( kind_ == KIND_QUERY )
        buf->appendf("     BPQ Query:\n");
    else if ( kind_ == KIND_RESPONSE )
        buf->appendf("   BPQ Response:\n");

    buf->appendf("Matching Rule: %d\n", matching_rule_);
    buf->appendf(" Query Length: %d\n", query_len_);
    buf->appendf("  Query Value: %s\n", query_val_);
    buf->appendf("\n");

}
*/
int
BPQBlock::write_to_buffer(u_char* buf, size_t len)
{
    int encoding_len=0; 
    u_int i=0, j=0;

    // BPQ-kind             1-byte
    if ( i < len ) 
        buf[i++] = (u_char) kind_;
    else
        return -1;

    // matching rule type   1-byte
    if ( i < len )
        buf[i++] = (u_char) matching_rule_;
    else
        return -1;

    // query-length         SDNV
    // todo: check this len -i is correct
    if ( i < len &&
        (encoding_len = SDNV::encode (query_len_, &(buf[i]), len -i)) >= 0 ) {
        i += encoding_len;
    } else {
        log_err_p(LOG, "Error encoding _BPQ query length");
        return -1;
    }

    // query-value           n-bytes
    for (j=0; query_val_ != NULL && i < len && j < query_len_; i++, j++)
        buf[i] = query_val_[j];    

    // todo: Still need to handle fragments
    if ( i < len &&
        (encoding_len = SDNV::encode (0, &(buf[i]), len -i)) >= 0 ) {
        i += encoding_len;
    } else {
        log_err_p(LOG, "Error encoding _BPQ fragment length");
        return -1;
    }

    return i;
}

u_int
BPQBlock::length() const
{
    // initial size {kind, matching rule}
    u_int len = 2; 
        
    len += SDNV::encoding_len(query_len_);
    len += query_len_;
    len += SDNV::encoding_len(0); // todo: frag len
    return len;
}

bool
BPQBlock::match(const BPQBlock* other) const
{
    return query_len_ == other->query_len() &&
           strncmp( (char*)query_val_, (char*)other->query_val(),
                     query_len_ ) == 0;
}

int
BPQBlock::initialise(BlockInfo* b)
{
    ASSERT ( b != NULL);

    BlockInfo* block = NULL;

    if ( b->source() != NULL ) {
        block = const_cast<BlockInfo*>(b->source());
        log_debug_p(LOG, "BPQBlock::initialise: b->source() != NULL");
    } else {
        log_debug_p(LOG, "BPQBlock::initialise: b->source() == NULL");
        block = b;
    }

    ASSERT ( block != NULL );
    
    log_debug_p(LOG, "BPQBlock::initialise: block != NULL");

    int decoding_len=0; 
    u_int i=0, j=0;
    u_int len = block->data_length();
    log_debug_p(LOG, "BPQBlock::initialise: len = %u", len);
    u_int num_frags;
    u_char* buf = block->data();
    log_debug_p(LOG, "BPQBlock::initialise: buf = %s", buf);

    // BPQ-kind             1-byte
    if ( i < len ) {
        kind_ = (kind_t) buf[i++];
        log_debug_p(LOG, "BPQBlock::initialise: kind = %d", kind_);
    }

    // matching rule type   1-byte
    if ( i < len ) {
        matching_rule_ = (u_int) buf[i++];
        log_debug_p(LOG, "BPQBlock::initialise: matching rule = %u", matching_rule_);
    }

    // Decode the SDNV-encoded query length. Note that we need to know the length of the
    // of the encoded value and provide some pointers to the encoded value along with
    // where we want the decoded value (in this case, query_len_).
    if ( i < len &&
        (decoding_len = SDNV::decode(&(buf[i]), len - i, &query_len_)) >= 0 ) {
        i += decoding_len;
        log_debug_p(LOG, "BPQBlock::initialise: query len = %u", query_len_);
    }
    else
        log_err_p(LOG, "Error decoding BPQ query length");

    // query-value           n-bytes
    if ( (i+query_len_) < len ) {
        query_val_ = (u_char*) malloc ( sizeof(u_char) * query_len_ );

        for (j=0; query_val_ != NULL && i < len && j < query_len_; i++, j++)
            query_val_[j] = buf[i];

        log_debug_p(LOG, "BPQBlock::initialise: query val = %s", query_val_);

    } else {
        query_val_ = NULL;
    }

    if ( i < len &&
        (decoding_len = SDNV::decode(&(buf[i]), len - i, &num_frags)) >= 0 ) {
        i += decoding_len;
        log_debug_p(LOG, "BPQBlock::initialise: num frags = %u", num_frags);
    }
    else
        log_err_p(LOG, "Error decoding BPQ fragment length");

    // todo: Still need to handle fragments
    // test assert - to be removed once we start handling fragments
    //ASSERT ( num_frags == 0 );
    if ( num_frags != 0 )
        log_err_p(LOG, "Error BPQ fragment length = %d", num_frags);

    return BP_SUCCESS;
}

} // namespace dtn

